/*****************************************************************************
  comon.h
  Last-modified: 21 Oct 2013 04:27:06 PM

  (c) 2012 - Yunfei Wang
  Center for Systems Biology
  Department of Molecular & Cell Biology
  University of Texas at Dallas
  tszn1984@gmail.com

  Licensed under the GNU General Public License 2.0 license.
******************************************************************************/

#ifndef COMON_H
#define COMON_H

// This is defined for wRNA g++ compile
#ifdef GPP_COMPILE
#define CPPEXTERN extern"C"
#else
#define CPPEXTERN extern
#endif

#endif //COMON_H

