#include "common.h"
/*

   energy constants, formerly defined in
             energy_par.h

   customized for use with RNAedit by
   S.Kopp, IMB-Jena, Germany, Mar 1996
   
*/

#ifndef _ENERGY_CONST_H
#define _ENERGY_CONST_H 1

#define GASCONST 1.98717  /* in [cal/K] */
#define K0  273.15
#define INF 1000000
#define FORBIDDEN 9999
#define BONUS 10000
#define NBPAIRS 7
#define TURN 3
#define MAXLOOP 30

#endif
